@@ -1,17 +1,18 @@
-from .util import *
+from .util import quote_string
 
-class Node(object):
+
+class Node:
     """
     A node within the garph.
     """
-    def __init__(self, node_id=None, alias=None, label=None, properties={}):
+    def __init__(self, node_id=None, alias=None, label=None, properties=None):
         """
         Create a new node
         """
         self.id = node_id
         self.alias = alias
         self.label = label
-        self.properties = properties
+        self.properties = properties or {}
 
     def toString(self):
         res = ""
@@ -36,8 +37,8 @@ class Node(object):
 
     def __eq__(self, rhs):
         # Quick positive check, if both IDs are set.
-        if self.id is not None and rhs.id is not None and self.id == rhs.id:
-            return True
+        if self.id is not None and rhs.id is not None and self.id != rhs.id:
+            return False
 
         # Label should match.
         if self.label != rhs.label:
