@@ -1,12 +1,16 @@
 import random
 import string
 
+__all__ = ['random_string', 'quote_string']
+
+
 def random_string(length=10):
     """
-    Returns a random N chracter long string.
+    Returns a random N character long string.
     """
     return ''.join(random.choice(string.ascii_lowercase) for x in range(length))
 
+
 def quote_string(v):
     """
     RedisGraph strings must be quoted,
@@ -21,6 +25,8 @@ def quote_string(v):
     if len(v) == 0:
         return '""'
 
+    v = v.replace('"', '\\"')
+
     if v[0] != '"':
         v = '"' + v
 
